﻿Imports Microsoft.VisualBasic
Imports System
' See "Learning Quickly When Irrelevant Attributes Abound: A New Linear-Threshold Algorithm"
' (1998), Nick Littlestone

Namespace WinnowPredict

    Friend Class WinnowProgram

        Shared Sub Main(ByVal args() As String)

            Console.WriteLine(vbLf & "Begin Winnow algorithm prediction demo")
            Console.WriteLine(vbLf & "Goal is predict political party based on 16 votes")

            Console.WriteLine(vbLf & "Example lines of the 435-item raw data file are:" & vbLf)
            Console.WriteLine("republican,n,y,n,y,y,y,n,n,n,n,n,y,y,y,n,?")
            Console.WriteLine("democrat,?,y,y,?,y,y,n,n,n,n,y,n,y,y,n,n")
            Console.WriteLine("democrat,n,y,y,n,y,y,n,n,n,n,n,n,y,y,y,y")

            Console.WriteLine(vbLf & "Loading hard-coded 100-item subset into memory")
            Console.WriteLine("All missing values ('?') replaced with 'n' (no)")
            Console.WriteLine("Moving political party to last column")
            Console.WriteLine("Encoding 'no' = 0, 'yes' = 1, 'dem' = 0, 'rep' = 1")

            Dim data(99)() As Integer
            data(0) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1} ' last col is dem = 0, rep = 1
            data(1) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(2) = New Integer() {0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0}
            data(3) = New Integer() {0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0}
            data(4) = New Integer() {1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0}
            data(5) = New Integer() {0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0}
            data(6) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0}
            data(7) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1}
            data(8) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1}
            data(9) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0}
            data(10) = New Integer() {0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1}
            data(11) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1}
            data(12) = New Integer() {0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0}
            data(13) = New Integer() {1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0}
            data(14) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1}
            data(15) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1}
            data(16) = New Integer() {1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0}
            data(17) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0}
            data(18) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1}
            data(19) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0}
            data(20) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0}
            data(21) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0}
            data(22) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0}
            data(23) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0}
            data(24) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0}
            data(25) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0}
            data(26) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0}
            data(27) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0}
            data(28) = New Integer() {1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1}
            data(29) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0}
            data(30) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(31) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0}
            data(32) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1, 1, 0}
            data(33) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1}
            data(34) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0}
            data(35) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(36) = New Integer() {1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1}
            data(37) = New Integer() {1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1}
            data(38) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(39) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0}
            data(40) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0}
            data(41) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0}
            data(42) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0}
            data(43) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0}
            data(44) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0}
            data(45) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0}
            data(46) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0}
            data(47) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}
            data(48) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0}
            data(49) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(50) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0}
            data(51) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1}
            data(52) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0}
            data(53) = New Integer() {1, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(54) = New Integer() {1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0}
            data(55) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1}
            data(56) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 1}
            data(57) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1}
            data(58) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1}
            data(59) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(60) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0}
            data(61) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(62) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0}
            data(63) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0}
            data(64) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0}
            data(65) = New Integer() {1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1}
            data(66) = New Integer() {0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1}
            data(67) = New Integer() {0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(68) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0}
            data(69) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0}
            data(70) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0}
            data(71) = New Integer() {1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 0, 1, 1}
            data(72) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0}
            data(73) = New Integer() {1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1}
            data(74) = New Integer() {1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0}
            data(75) = New Integer() {0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0}
            data(76) = New Integer() {0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0}
            data(77) = New Integer() {0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0}
            data(78) = New Integer() {1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0}
            data(79) = New Integer() {0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(80) = New Integer() {1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0}
            data(81) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0}
            data(82) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1}
            data(83) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(84) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(85) = New Integer() {0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0}
            data(86) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}
            data(87) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1}
            data(88) = New Integer() {0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0}
            data(89) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1}
            data(90) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0}
            data(91) = New Integer() {1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0}
            data(92) = New Integer() {1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0}
            data(93) = New Integer() {1, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 0}
            data(94) = New Integer() {1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0}
            data(95) = New Integer() {1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0}
            data(96) = New Integer() {0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0}
            data(97) = New Integer() {1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0}
            data(98) = New Integer() {1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0}
            data(99) = New Integer() {0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1}

            Console.WriteLine(vbLf & "First few lines of all data are: " & vbLf)
            ShowMatrix(data, 4, True)

            Console.WriteLine(vbLf & "Splitting data into 80% train" & " and 20% test matrices")
            Dim trainData()() As Integer = Nothing
            Dim testData()() As Integer = Nothing
            MakeTrainTest(data, 0, trainData, testData)

            Console.WriteLine(vbLf & "Encoding 'n' and '?' = 0, 'y' = 1, 'democrat' = 0, 'republican' = 1")
            Console.WriteLine("Moving political party to last column")
            Console.WriteLine(vbLf & "First few rows of training data are:" & vbLf)
            ShowMatrix(trainData, 3, True)

            Console.WriteLine(vbLf & "Begin training using Winnow algorithm")
            Dim numInput As Integer = 16
            Dim w As New Winnow(numInput, 0) ' rndSeed = 0
            Dim weights() As Double = w.Train(trainData)
            Console.WriteLine("Training complete")

            Console.WriteLine(vbLf & "Final model weights are:")
            ShowVector(weights, 4, 8, True)

            Dim trainAcc As Double = w.Accuarcy(trainData)
            Dim testAcc As Double = w.Accuarcy(testData)

            Console.WriteLine(vbLf & "Prediction accuracy on training data = " & trainAcc.ToString("F4"))
            Console.WriteLine("Prediction accuracy on test data = " & testAcc.ToString("F4"))

            Console.WriteLine(vbLf & "Predicting party of Representative with all 'yes' votes")
            Dim unknown() As Integer = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}
            Dim predicted As Integer = w.ComputeOutput(unknown)
            If predicted = 0 Then
                Console.WriteLine("Prediction is 'democrat'")
            Else
                Console.WriteLine("Prediction is 'republican'")
            End If

            Console.WriteLine(vbLf & "End Winnow demo" & vbLf)
            Console.ReadLine()

        End Sub ' Main

        Private Shared Sub MakeTrainTest(ByVal data()() As Integer, ByVal seed As Integer, <System.Runtime.InteropServices.Out()> ByRef trainData()() As Integer, <System.Runtime.InteropServices.Out()> ByRef testData()() As Integer)

            Dim rnd As New Random(seed)
            Dim totRows As Integer = data.Length ' compute number of rows in each result
            Dim numTrainRows As Integer = CInt(Fix(totRows * 0.8))
            Dim numTestRows As Integer = totRows - numTrainRows
            trainData = New Integer(numTrainRows - 1)() {}
            testData = New Integer(numTestRows - 1)() {}

            Dim copy(data.Length - 1)() As Integer ' make a copy of data
            For i As Integer = 0 To copy.Length - 1 ' by reference to save space
                copy(i) = data(i)
            Next i
            For i As Integer = 0 To copy.Length - 1 ' scramble row order of copy

                Dim r As Integer = rnd.Next(i, copy.Length)
                Dim tmp() As Integer = copy(r)
                copy(r) = copy(i)
                copy(i) = tmp

            Next i
            For i As Integer = 0 To numTrainRows - 1 ' create training
                trainData(i) = copy(i)
            Next i

            For i As Integer = 0 To numTestRows - 1 ' create test
                testData(i) = copy(i + numTrainRows)
            Next i

        End Sub ' MakeTrainTest

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal valsPerRow As Integer, ByVal newLine As Boolean)

            For i As Integer = 0 To vector.Length - 1

                If i Mod valsPerRow = 0 Then
                    Console.WriteLine("")
                End If
                Console.Write(vector(i).ToString("F" & decimals).PadLeft(decimals + 4) & " ")

            Next i
            If newLine = True Then
                Console.WriteLine("")
            End If

        End Sub

        Private Shared Sub ShowMatrix(ByVal matrix()() As Integer, ByVal numRows As Integer, ByVal indices As Boolean)

            For i As Integer = 0 To numRows - 1

                If indices = True Then
                    Console.Write("[" & i.ToString().PadLeft(2) & "]   ")
                End If
                For j As Integer = 0 To matrix(i).Length - 1

                    Console.Write(matrix(i)(j) & " ")

                Next j
                Console.WriteLine("")

            Next i
            Dim lastIndex As Integer = matrix.Length - 1
            If indices = True Then
                Console.Write("[" & lastIndex.ToString().PadLeft(2) & "]   ")
            End If
            For j As Integer = 0 To matrix(lastIndex).Length - 1
                Console.Write(matrix(lastIndex)(j) & " ")
            Next j
            Console.WriteLine(vbLf)

        End Sub

    End Class ' Program

    Public Class Winnow

        Private numInput As Integer
        Private weights() As Double
        Private threshold As Double ' to determine Y = 0 or 1
        Private alpha As Double ' increase/decrase factor
        Private Shared rnd As Random

        Public Sub New(ByVal numInput As Integer, ByVal rndSeed As Integer)

            Me.numInput = numInput
            Me.weights = New Double(numInput - 1) {}
            For i As Integer = 0 To weights.Length - 1
                weights(i) = numInput / 2.0
            Next i
            Me.threshold = 1.0 * numInput
            Me.alpha = 2.0
            rnd = New Random(rndSeed)

        End Sub

        Public Function ComputeOutput(ByVal xValues() As Integer) As Integer

            Dim sum As Double = 0.0
            For i As Integer = 0 To numInput - 1
                sum += weights(i) * xValues(i)
            Next i
            If sum > Me.threshold Then
                Return 1
            Else
                Return 0
            End If

        End Function

        Public Function Train(ByVal trainData()() As Integer) As Double()

            Dim xValues(numInput - 1) As Integer
            Dim target As Integer
            Dim computed As Integer
            Shuffle(trainData)
            For i As Integer = 0 To trainData.Length - 1

                Array.Copy(trainData(i), xValues, numInput) ' get the inputs
                target = trainData(i)(numInput) ' last value is target
                computed = ComputeOutput(xValues)

                If computed = 1 AndAlso target = 0 Then ' need to decrease weights

                    For j As Integer = 0 To numInput - 1

                        If xValues(j) = 0 Then ' no change when xi = 0
                            Continue For
                        End If
                        weights(j) = weights(j) / alpha ' demotion

                    Next j

                ElseIf computed = 0 AndAlso target = 1 Then ' need to increase weights

                    For j As Integer = 0 To numInput - 1

                        If xValues(j) = 0 Then ' no change when xi = 0
                            Continue For
                        End If
                        weights(j) = weights(j) * alpha ' promotion

                    Next j

                End If

            Next i ' each training item

            Dim result(numInput - 1) As Double ' = number weights
            Array.Copy(Me.weights, result, numInput)
            Return result

        End Function ' Train

        Private Shared Sub Shuffle(ByVal trainData()() As Integer) ' Fisher-Yates shuffle algorithm

            For i As Integer = 0 To trainData.Length - 1

                Dim r As Integer = rnd.Next(i, trainData.Length)
                Dim tmp() As Integer = trainData(r)
                trainData(r) = trainData(i)
                trainData(i) = tmp

            Next i

        End Sub

        Public Function Accuarcy(ByVal trainData()() As Integer) As Double

            Dim numCorrect As Integer = 0
            Dim numWrong As Integer = 0

            Dim xValues(numInput - 1) As Integer
            Dim target As Integer
            Dim computed As Integer

            For i As Integer = 0 To trainData.Length - 1

                Array.Copy(trainData(i), xValues, numInput) ' get the inputs
                target = trainData(i)(numInput) ' last value is target
                computed = ComputeOutput(xValues)

                If computed = target Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If

            Next i
            Return (numCorrect * 1.0) / (numCorrect + numWrong)

        End Function

    End Class ' Winnow

End Namespace ' ns